
//
// PulseTLM.c
//
// This is PulseTLM a more full-featured Transmission Line Matrix
// (TLM) method simulation engine.
//
// Copyright (C) 1999 Paul Hayes
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or any later version, with the following conditions
// attached in addition to any and all conditions of the GNU
//
// General Public License: When reporting or displaying any results or
// animations created using this code or modification of this code,
// make the appropriate citation referencing PulseTLM by name and
// including the version number.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// Contacting the authors:
//
// Paul Hayes
//
// info@cemtach.org
// 
// http://cemtach.org
// http://cemtach.org/software/software.html
// http://cemtach.org/software/PulseTLM/PulseTLM.html
//

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
   #include <unistd.h>
#endif

#include "PulseTLM.h"






void
parseFile(FILE *inputFile,
	  FILE *logFile,
	  int *nx,
	  int *ny,
	  int *nz,
	  int *maximumIteration,
	  double *maximumSimulationTime,
	  Boundary *boundaryList,
	  Boundary *lastBoundary,
	  Excitation *excitationList,
	  Excitation *lastExcitation,
	  Material *materialList,
	  Material *lastMaterial,
	  Output *outputList,
	  Output *lastOutput,
	  double *dx,
	  double *dy,
	  double *dz)
{
  //
  // An error string for general otuput of errors and status messages
  //
  char errorString[1024];
  //
  // Input line.  Each line of the input file must be shorter than this.
  //
  char inputLine[1024];
  //
  // Input line type as the first space separated string of an input line
  //
  char inputLineType[1024];
  //
  // Type and field parsing variables
  //
  char type[1024], field[1024];
  //
  // The current material as the last and list were passed into this
  // function since they will be usefull in the main code.
  //
  Material *material = NULL;
  //
  // The current boundary as the last and list were passed into this
  // function since they will be usefull in the main code.
  //
  Boundary *boundary = NULL;
  //
  // The current excitation as the last and list were passed into this
  // function since they will be usefull in the main code.
  //
  Excitation *excitation = NULL;
  //
  // The current output as the last and list were passed into this
  // function since they will be usefull in the main code.
  //
  Output *output = NULL;
  //
  // Generic array traversal variables
  //
  int i,j,k,n;
  //
  // Count the number of errors.  If it is greater than zero at the
  // end of the parse, then exit suggesting that the user work over
  // the input file.
  //
  int numberOfErrors = 0;
  //
  //
  //
  //
  // Suck in each inputline one at a time and parse into the TLM mesh.
  // Note that this requires that mesh control lines appear at the
  // beginning of the input file.
  //
  while(fgets(inputLine, 1024, inputFile) != NULL)
    {
    //
    // Clear the input line type before each parse
    //
    strcpy(inputLineType,"");
    //
    // Pick the input line type out from the input line as the first
    // space separated string
    //
    sscanf(inputLine,"%s", inputLineType);
    if(strcmp(inputLineType, "mesh") == 0)
      {
      //
      // The mesh spec line has been located: set nx, ny and nz
      //
      sscanf(inputLine, "%s%d%d%d", inputLineType, nx, ny, nz);
      fprintf(logFile, "mesh %d %d %d\n", *nx, *ny, *nz);
      //
      // Each mesh dimension must be greater than 1 as that specified
      // a planar aspect in that coordinate direction.
      //
      if (*nx < 1)
	{
	sprintf(errorString, "nx must be > 1\n");
	fprintf(logFile, errorString);
	numberOfErrors++;
	}
      if (*ny < 1)
	{
	sprintf(errorString, "ny must be > 1\n");
	fprintf(logFile, errorString);
	numberOfErrors++;
	}
      if (*nz < 1)
	{
	sprintf(errorString, "nz must be > 1\n");
	fprintf(logFile, errorString);
	numberOfErrors++;
	}
      }
    else if(strcmp(inputLineType,"maximumIteration")==0)
      {
      //
      // The maximumIteration line has been located: set it
      //
      sscanf(inputLine, "%s%d", inputLineType, maximumIteration);
      fprintf(logFile, "maximumIteration %d\n", *maximumIteration);
      }
    else if(strcmp(inputLineType,"maximumSimulationTime")==0)
      {
      //
      // The maximumSimulationTime line has been located: set it
      //
      sscanf(inputLine, "%s%lg", inputLineType, maximumSimulationTime);
      fprintf(logFile, "maximumSimulationTime %lg\n", *maximumSimulationTime);
      }
    else if(strcmp(inputLineType,"material") == 0)
      {
      //
      // A media block has been encountred in the input file, add the
      // media line to the current list of media lines.
      //
      if (materialList == NULL)
	{
	//
	// This is the first media line within this input file, so
	// start the list setting material to be the one about to be
	// filled out and lastMaterial as the last in the list.
	//
	materialList = (Material *)malloc(sizeof(Material));
	material = materialList;
	}
      else
	{
	//
	// This is not the first media line, to continue the list at
	// the end setting material to be the one about to be filled
	// out and lastMaterial as the last in the list.
	//
	lastMaterial->next = (Material *)malloc(sizeof(Material));
	material = lastMaterial->next;
	}
      //
      // Truncate the list
      //
      material->next = NULL;
      lastMaterial = material;
      //
      // Parse out the information for this media block
      //
      sscanf(inputLine, "%s%d%d%d%d%d%d%lf%lf%lf",
	     inputLineType,
	     &(material->xLow),
	     &(material->xHigh),
	     &(material->yLow),
	     &(material->yHigh),
	     &(material->zLow),
	     &(material->zHigh),
	     &(material->relativePermeability),
	     &(material->relativePermittivity),
	     &(material->conductivity));
      //
      // Echo it to the logFile for user feedback
      //
      fprintf(logFile, "material %d %d   %d %d   %d %d %lg %lg %lg\n",
	      material->xLow,
	      material->xHigh,
	      material->yLow,
	      material->yHigh,
	      material->zLow,
	      material->zHigh,
	      material->relativePermeability,
	      material->relativePermittivity,
	      material->conductivity);
      }
    else if(strcmp(inputLineType,"rtxl") == 0 ||
	    strcmp(inputLineType,"rtyl") == 0 ||
	    strcmp(inputLineType,"rtzl") == 0 ||
	    strcmp(inputLineType,"rtxh") == 0 ||
	    strcmp(inputLineType,"rtyh") == 0 ||
	    strcmp(inputLineType,"rtzh") == 0)
      {
      //
      // A reflective-transmissive boundary has been encountered in
      // the input file, add to the list of boundary conditions.
      //
      if (boundaryList->type == 0)
	{
	//
	// This is the initial `key' boundary in the overall list
	// simply fill it out since it was allocated by the main code.
	//
	boundary = boundaryList;
	}
      else
	{
	//
	// This is not the first boundary, so allocate and add to the
	// end of the list setting boundary to be the one about to be
	// filled out and lastBoundary as the last in the list.
	//
	lastBoundary->next = (Boundary *)malloc(sizeof(Boundary));
	boundary = lastBoundary->next;
	}
      //
      // Truncate the list
      //
      boundary->next = NULL;
      lastBoundary = boundary;
      //
      // Parse out the reflective transmissive boundary condition
      //
      n = sscanf(inputLine, "%s%d%d%d%d%d%d%lf%lf", 
		 inputLineType, 
		 &(boundary->xLow), 
		 &(boundary->xHigh), 
		 &(boundary->yLow), 
		 &(boundary->yHigh), 
		 &(boundary->zLow), 
		 &(boundary->zHigh), 
		 &(boundary->rho),
		 &(boundary->tau));
      //
      // Fill in tau if the user didn't explicitliy specify it 
      //
      if (n < 9)
	{
	boundary->tau = 1.0 + boundary->rho;
	}
      //
      // Determine the int type of the boundary
      //
      if (strcmp(inputLineType, "rtxh") == 0)
	{
	boundary->type = 1;
	}
      if (strcmp(inputLineType, "rtyh") == 0)
	{
	boundary->type = 2;
	}
      if (strcmp(inputLineType, "rtzh") == 0)
	{
	boundary->type = 3;
	}
      if (strcmp(inputLineType, "rtxl") == 0)
	{
	boundary->type = 4;
	}
      if (strcmp(inputLineType, "rtyl") == 0)
	{
	boundary->type = 5;
	}
      if (strcmp(inputLineType, "rtzl") == 0)
	{
	boundary->type = 6;
	}
      //
      // Echo the boundary to the logFile
      //
      fprintf(logFile,"%s %d %d   %d %d   %d %d   %lg %lg\n",
	      inputLineType,
	      boundary->xLow,
	      boundary->xHigh,
	      boundary->yLow,
	      boundary->yHigh,
	      boundary->zLow,
	      boundary->zHigh,
	      boundary->rho,
	      boundary->tau);
      //
      // Check for obvious errors
      //
      if ( (boundary->rho < -1.0) || (boundary->rho > 1.0) )
	{
	sprintf(errorString, "Flaw in rho/tau of the boundary");
	fprintf(logFile, errorString);
	numberOfErrors++;
	}
      }
    else if(strcmp(inputLineType, "dx") == 0)
      {
      //
      // Parse the x spatial delta from the file
      //
      sscanf(inputLine,"%s%lf",
	     inputLineType,
	     dx);
      //
      // Echo the x spatial delta to the logfile
      //
      fprintf(logFile, "dx %lf\n", *dx);
      }
    else if(strcmp(inputLineType, "dy") == 0)
      {
      //
      // Parse the y spatial delta from the file
      //
      sscanf(inputLine,"%s%lf",
	     inputLineType,
	     dy);
      //
      // Echo the y spatial delta to the logfile
      //
      fprintf(logFile, "dy %lf\n", *dy);
      }
    else if(strcmp(inputLineType, "dz") == 0)
      {
      //
      // Parse the z spatial delta from the file
      //
      sscanf(inputLine,"%s%lf",
	     inputLineType,
	     dz);
      //
      // Echo the z spatial delta to the logfile
      //
      fprintf(logFile, "dz %lf\n", *dz);
      }
    else if(strcmp(inputLineType, "excitation") == 0)
      {
      //
      // An excitation has been encountered in the input file, add to the
      // list of excitation conditions.
      //
      if (excitationList->type == 0)
	{
	//
	// This is the initial `key' excitation in the overall list
	// simply fill it out since it was allocated by the main code.
	//
	excitation = excitationList;
	}
      else
	{
	//
	// This is not the first excitation, so allocate and add to the
	// end of the list setting excitation to be the one about to be
	// filled out and lastExcitation as the last in the list.
	//
	lastExcitation->next = (Excitation *)malloc(sizeof(Excitation));
	excitation = lastExcitation->next;
	}
      //
      // Truncate the list
      //
      excitation->next = NULL;
      lastExcitation = excitation;
      //
      // Force initial conditions on the values
      //
      excitation->value1 = 0.0;
      excitation->value2 = 0.0;
      excitation->value3 = 0.0;
      excitation->value4 = 0.0;
      excitation->value5 = 0.0;
      //
      // Parse out the excitation
      //
      n = sscanf(inputLine, "%s%d%d%d%d%d%d%s%s%lf%lf%lf%lf%lf", 
		 inputLineType, 
		 &(excitation->xLow), 
		 &(excitation->xHigh), 
		 &(excitation->yLow), 
		 &(excitation->yHigh), 
		 &(excitation->zLow), 
		 &(excitation->zHigh),
		 field,
		 type,
		 &(excitation->value1),
		 &(excitation->value2),
		 &(excitation->value3),
		 &(excitation->value4),
		 &(excitation->value5));
      //
      // Determine the int type of the field
      //
      if (strcmp(field, "ex") == 0)
	{
	excitation->field = 1;
	}
      if (strcmp(field, "ey") == 0)
	{
	excitation->field = 2;
	}
      if (strcmp(field, "ez") == 0)
	{
	excitation->field = 3;
	}
      if (strcmp(field, "hx") == 0)
	{
	excitation->field = 4;
	}
      if (strcmp(field, "hy") == 0)
	{
	excitation->field = 5;
	}
      if (strcmp(field, "hz") == 0)
	{
	excitation->field = 6;
	}
      //
      // Determine the int type of the excitation type
      //
      if (strcmp(type, "impulse") == 0)
	{
	excitation->type = 1;
	}
      if (strcmp(type, "sin") == 0)
	{
	excitation->type = 2;
	}
      if (strcmp(type, "exp1") == 0)
	{
	excitation->type = 3;
	}
      //
      // Echo out the parsed information to the logFile in a way that
      // confirms the parse was performed correctly.
      //
      fprintf(logFile, "excitation %d %d   %d %d   %d %d ",
	      excitation->xLow,
	      excitation->xHigh,
	      excitation->yLow,
	      excitation->yHigh,
	      excitation->zLow,
	      excitation->zHigh);
      //
      if (excitation->field == 1)
	{
	fprintf(logFile, "ex");
	}
      if (excitation->field == 2)
	{
	fprintf(logFile, "ey");
	}
      if (excitation->field == 3)
	{
	fprintf(logFile, "ez");
	}
      if (excitation->field == 4)
	{
	fprintf(logFile, "hx");
	}
      if (excitation->field == 5)
	{
	fprintf(logFile, "hy");
	}
      if (excitation->field == 6)
	{
	fprintf(logFile, "hz");
	}
      //
      if (excitation->type == 1)
	{
	fprintf(logFile, " impulse %lg", excitation->value1);
	}
      if (excitation->type == 2)
	{
	fprintf(logFile, " %lg * sin(2.0*M_PI*%lg*(t - %lg)) + %lg",
		excitation->value1,
		excitation->value2,
		excitation->value3,
		excitation->value4);
	}
      if (excitation->type == 3)
	{
	fprintf(logFile, " %lg * exp( -pow((t - %lg)/%lg, 2.0) ) + %lg",
		excitation->value1,
		excitation->value2,
		excitation->value3,
		excitation->value4);
	}
      fprintf(logFile, "\n");
      }
    else if(strcmp(inputLineType, "output") == 0)
      {
      //
      // An output has been encountered in the input file, add to the
      // list of output conditions.
      //
      if (outputList->type == 0)
	{
	//
	// This is the initial `key' output in the overall list
	// simply fill it out since it was allocated by the main code.
	//
	output = outputList;
	}
      else
	{
	//
	// This is not the first output, so allocate and add to the
	// end of the list setting output to be the one about to be
	// filled out and lastOutput as the last in the list.
	//
	lastOutput->next = (Output *)malloc(sizeof(Output));
	output = lastOutput->next;
	}
      //
      // Truncate the list
      //
      output->next = NULL;
      lastOutput = output;
      //
      // Parse out the excitation
      //
      n = sscanf(inputLine, "%s%d%d%d%d%d%d%s%s%d", 
		 inputLineType, 
		 &(output->xLow), 
		 &(output->xHigh), 
		 &(output->yLow), 
		 &(output->yHigh), 
		 &(output->zLow), 
		 &(output->zHigh),
		 field,
		 type,
		 &(output->modulo));
      //
      // Determine the int type of the field
      //
      if (strcmp(field, "ex") == 0)
	{
	output->field = 1;
	}
      if (strcmp(field, "ey") == 0)
	{
	output->field = 2;
	}
      if (strcmp(field, "ez") == 0)
	{
	output->field = 3;
	}
      if (strcmp(field, "hx") == 0)
	{
	output->field = 4;
	}
      if (strcmp(field, "hy") == 0)
	{
	output->field = 5;
	}
      if (strcmp(field, "hz") == 0)
	{
	output->field = 6;
	}
      //
      if (strcmp(field, "eMagnitude") == 0)
	{
	output->field = 7;
	}
      if (strcmp(field, "hMagnitude") == 0)
	{
	output->field = 8;
	}
      //
      // Determine the int type of the output type
      //
      if (strcmp(type, "viz") == 0)
	{
	output->type = 1;
	}
      if (strcmp(type, "pnm") == 0)
	{
	output->type = 2;
	}
      if (strcmp(type, "mtv") == 0)
	{
	output->type = 3;
	}
      if (strcmp(type, "grace") == 0)
	{
	output->type = 4;
	}
      if (strcmp(type, "dat") == 0)
	{
	output->type = 5;
	}
      if (strcmp(type, "binary") == 0)
	{
	output->type = 6;
	}
      //
      // Need to do input checking here.
      //
      }
    else
      {
      //
      // The line is inrecognized but echo it to the logfile so the
      // user knows it was unrecognized.
      //
      fprintf(logFile, "##%s", inputLine);
      }
    
    }
  //
  // Each of the spatial deltas must have been set
  //
  if (*dx <= 0.0)
    {
    sprintf(errorString, "dx must be set > 0.0.  Currently: dx = %lf\n", *dx);
    fprintf(logFile, errorString);
    numberOfErrors++;
    }
  if (*dy <= 0.0)
    {
    sprintf(errorString, "dy must be set > 0.0.  Currently: dy = %lf\n", *dy);
    fprintf(logFile, errorString);
    numberOfErrors++;
    }
  if (*dz <= 0.0)
    {
    sprintf(errorString, "dz must be set > 0.0.  Currently: dz = %lf\n", *dz);
    fprintf(logFile, errorString);
    numberOfErrors++;
    }
  //
  // If neither maximumIteration nor maximumSimulationTime have been
  // set, then this simulation really won't run, so error for the user
  // to fix.
  //
  if ( (*maximumIteration <= 0) && (*maximumSimulationTime <= 0.0) )
    {
    sprintf(errorString, "One of maximumIteration or maximumSimulationTime must be set\n");
    fprintf(logFile, errorString);
    numberOfErrors++;
    }
  //
  // Traverse the outputs and number them uniquely which helps keep
  // separate output requests separate.
  //
  output = outputList;
  n = 0;
  while(output != NULL)
    {
    output->uniqueNumber = n;
    n++;
    output = output->next;
    }
  //
  // Traverse the excitations and number them uniquely which helps
  // keep separate output requests separate.
  //
  excitation = excitationList;
  n = 0;
  while(excitation != NULL)
    {
    excitation->uniqueNumber = n;
    n++;
    excitation = excitation->next;
    }
  //
  // If any errors have occured, exit to allow the user to fix the
  // input file.
  //
  if (numberOfErrors > 0)
    {
    sprintf(errorString, "Errors have occured in parsing: Simulation Stopped.\n");
    fprintf(logFile, errorString);
    exit(9);
    }
  //
  // Finish the input file parsing by removing any unnecessary information
  //
}
